#include "stdafx.h"
#include <winioctl.h>
#include <ntddscsi.h>
#include <stdio.h>
#include "wnaspi32.h"
#include "scsidefs.h"
#include "ASPTI.h"

#define offsetof(s,m)   (size_t)&(((s *)0)->m)
#define SIZE_BUFFER 0x10000
#define DEVICE_MAX ('Z'-'A'+1)
#define BUS_WIDTH 32

int g_deviceNumber;
BYTE g_aDevice[DEVICE_MAX];
char g_aNameDrive[DEVICE_MAX][16];
BYTE *g_pBuffer;

BOOL APIENTRY DllMain(HANDLE hModule, DWORD reason, LPVOID)
{
	if (DLL_PROCESS_ATTACH == reason)
	{
		scanBus();
		g_pBuffer = (BYTE *)VirtualAlloc(NULL, SIZE_BUFFER, MEM_COMMIT| MEM_RESERVE| MEM_TOP_DOWN, PAGE_READWRITE| PAGE_NOCACHE);
	}
	else if (DLL_PROCESS_DETACH == reason)
	{
		if (NULL != g_pBuffer)
			VirtualFree(g_pBuffer, 0, MEM_RELEASE);
	}

	return TRUE;
}

__declspec(dllexport) DWORD	__cdecl GetASPI32SupportInfo()
{
	DWORD returnValue;


	scanBus();
	returnValue = (SS_COMP<<8)| (1&0xff);

	return returnValue;
}

__declspec(dllexport) DWORD	__cdecl SendASPI32Command(LPSRB pSrb)
{
	DWORD returnValue;
	PSRB_Header pHeader;


	returnValue = SS_ERR;
	pHeader = PSRB_Header(pSrb);
	switch (pHeader->SRB_Cmd)
	{
	case SC_HA_INQUIRY:
		returnValue = command_HaInquiry(pSrb);
		break;

	case SC_GET_DEV_TYPE:
		returnValue = command_GetDevType(pSrb);
		break;

	case SC_EXEC_SCSI_CMD:
		returnValue = command_ExecScsiCmd(pSrb);
		break;

	case SC_ABORT_SRB:
		returnValue = command_AbortSrb(pSrb);
		break;

	case SC_RESET_DEV:
		returnValue = command_ResetDev(pSrb);
		break;

	case SC_SET_HA_PARMS:
		returnValue = SS_INVALID_CMD;
		break;

	case SC_GET_DISK_INFO:
		returnValue = command_GetDiskInfo(pSrb);
		break;

	case SC_RESCAN_SCSI_BUS:
		returnValue = command_RescanScsiBus(pSrb);
		break;

	case SC_GETSET_TIMEOUTS:
		returnValue = command_GetsetTimeouts(pSrb);
		break;

	default:
		returnValue = SS_INVALID_CMD;
		break;
	}

	return returnValue;
}


//
void scanBus()
{
	int i;
	char aNameDrive[] = "A:\\";


	for (i=g_deviceNumber=0; i<='Z'-'A'; i++)
	{
		sprintf(aNameDrive, "%c:\\", 'A'+i);
		if (DRIVE_CDROM == GetDriveType(aNameDrive))
		{
			g_aDevice[g_deviceNumber] = BYTE(i);
			sprintf(g_aNameDrive[g_deviceNumber++], "\\\\.\\%c:", 'A'+i);
		}
	}
}


DWORD command_HaInquiry(LPSRB pSrb_)
{
	PSRB_HAInquiry pSrb;
	static const char aManagerId[] = "ASPTI for WinNT ";
	static const char aHostAdapterId[] = "virtual SCSI-HA ";


	pSrb = PSRB_HAInquiry(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	pSrb->HA_Count = 1;
	CopyMemory(pSrb->HA_ManagerId, aManagerId, 16);
	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	pSrb->SRB_Status = SS_COMP;
	pSrb->HA_SCSI_ID = BUS_WIDTH-1;
	CopyMemory(pSrb->HA_Identifier, aHostAdapterId, 16);
	ZeroMemory(pSrb->HA_Unique, sizeof pSrb->HA_Unique);
	pSrb->HA_Unique[3] = BUS_WIDTH;
	*(DWORD *)(&pSrb->HA_Unique[4]) = SIZE_BUFFER;

	return pSrb->SRB_Status;
}

DWORD command_GetDevType(LPSRB pSrb_)
{
	PSRB_GDEVBlock pSrb;


	pSrb = PSRB_GDEVBlock(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	if (g_deviceNumber<=pSrb->SRB_Target || pSrb->SRB_Lun)
		return pSrb->SRB_Status = SS_NO_DEVICE;

	pSrb->SRB_Status = SS_COMP;
	pSrb->SRB_DeviceType = DTYPE_CDROM;

	return pSrb->SRB_Status;
}

DWORD command_ExecScsiCmd(LPSRB pSrb_)
{
	PSRB_ExecSCSICmd pSrb;
	SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER sptBlock;
	HANDLE hDrive;
	DWORD length, tmp;


	pSrb = PSRB_ExecSCSICmd(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	if (g_deviceNumber<=pSrb->SRB_Target || pSrb->SRB_Lun)
		return pSrb->SRB_Status = SS_NO_DEVICE;

	ZeroMemory(&sptBlock, sizeof SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER);
	sptBlock.sptd.Length = sizeof SCSI_PASS_THROUGH_DIRECT;
	sptBlock.sptd.CdbLength = pSrb->SRB_CDBLen;
	sptBlock.sptd.SenseInfoLength = pSrb->SRB_SenseLen;
	if (SRB_DIR_IN & pSrb->SRB_Flags)
		sptBlock.sptd.DataIn |= SCSI_IOCTL_DATA_IN;
	else if (SRB_DIR_OUT & pSrb->SRB_Flags)
		sptBlock.sptd.DataIn |= SCSI_IOCTL_DATA_OUT;
	else
		sptBlock.sptd.DataIn |= SCSI_IOCTL_DATA_UNSPECIFIED;

	sptBlock.sptd.DataTransferLength = pSrb->SRB_BufLen;
	sptBlock.sptd.TimeOutValue = 0xffff;
	sptBlock.sptd.DataBuffer = g_pBuffer;
	sptBlock.sptd.SenseInfoOffset = offsetof(SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER, ucSenseBuf);
	CopyMemory(sptBlock.sptd.Cdb, pSrb->CDBByte, 16);

	hDrive = CreateFile(g_aNameDrive[pSrb->SRB_Target], GENERIC_READ| GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (INVALID_HANDLE_VALUE == hDrive)
		return pSrb->SRB_Status = SS_NO_DEVICE;

	length = sizeof SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER;
	if (!DeviceIoControl(hDrive, IOCTL_SCSI_PASS_THROUGH_DIRECT, &sptBlock, length, &sptBlock, length, &tmp, NULL))
	{
		CloseHandle(hDrive);
		return pSrb->SRB_Status = SS_ERR;
	}

	CloseHandle(hDrive);
	pSrb->SRB_Status = SS_COMP;
	CopyMemory(pSrb->SRB_BufPointer, g_pBuffer, pSrb->SRB_BufLen);
	pSrb->SRB_HaStat = HASTAT_OK;
	pSrb->SRB_TargStat = STATUS_GOOD;
	CopyMemory(pSrb->SenseArea, sptBlock.ucSenseBuf, pSrb->SRB_SenseLen);
	if (SRB_EVENT_NOTIFY & pSrb->SRB_Flags)
		SetEvent(HANDLE(pSrb->SRB_PostProc));
	else if (SRB_POSTING & pSrb->SRB_Flags)
		((void (__cdecl *)(PSRB_ExecSCSICmd))pSrb->SRB_PostProc)(pSrb);

	return pSrb->SRB_Status;
}

DWORD command_AbortSrb(LPSRB pSrb_)
{
	PSRB_Abort pSrb;


	pSrb = PSRB_Abort(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	pSrb->SRB_Status = SS_COMP;

	return pSrb->SRB_Status;
}

DWORD command_ResetDev(LPSRB pSrb_)
{
	PSRB_BusDeviceReset pSrb;


	pSrb = PSRB_BusDeviceReset(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	if (g_deviceNumber<=pSrb->SRB_Target || pSrb->SRB_Lun)
		return pSrb->SRB_Status = SS_NO_DEVICE;

	pSrb->SRB_Status = SS_COMP;

	return pSrb->SRB_Status;
}

DWORD command_GetDiskInfo(LPSRB pSrb_)
{
	PSRB_GetDiskInfo pSrb;


	pSrb = PSRB_GetDiskInfo(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	if (g_deviceNumber<=pSrb->SRB_Target || pSrb->SRB_Lun)
		return pSrb->SRB_Status = SS_NO_DEVICE;

	pSrb->SRB_Status = SS_COMP;
	pSrb->SRB_DriveFlags = 0; //DISK_NOT_INT13

	return pSrb->SRB_Status;
}

DWORD command_RescanScsiBus(LPSRB pSrb_)
{
	PSRB_RescanPort pSrb;


	pSrb = PSRB_RescanPort(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	pSrb->SRB_Status = SS_COMP;
	scanBus();

	return pSrb->SRB_Status;
}

DWORD command_GetsetTimeouts(LPSRB pSrb_)
{
	PSRB_GetSetTimeouts pSrb;


	pSrb = PSRB_GetSetTimeouts(pSrb_);
	if (NULL == pSrb)
		return SS_INVALID_SRB;

	if (0 != pSrb->SRB_HaId)
		return pSrb->SRB_Status = SS_INVALID_HA;

	if (SRB_DIR_IN == pSrb->SRB_Flags)
	{
		if (g_deviceNumber<=pSrb->SRB_Target || pSrb->SRB_Lun)
			return pSrb->SRB_Status = SS_INVALID_SRB;

		pSrb->SRB_Timeout = DEFAULT_SRB_TIMEOUT;
	}
	else if (SRB_DIR_OUT == pSrb->SRB_Flags)
	{
		if ((g_deviceNumber<=pSrb->SRB_Target && 0xff!=pSrb->SRB_Target) ||
			(0!=pSrb->SRB_Lun && 0xff!=pSrb->SRB_Lun))
			return pSrb->SRB_Status = SS_INVALID_SRB;
	}
	else
		return pSrb->SRB_Status = SS_INVALID_SRB;

	pSrb->SRB_Status = SS_COMP;

	return pSrb->SRB_Status;
}



